"""Test winmsvcr."""

from __future__ import annotations

import sys

import pytest

from freeze_core._compat import IS_LINUX, IS_WINDOWS
from freeze_core.winmsvcr import MSVC_FILES, UCRT_FILES

# This test is really necessary on Windows, but it runs in other environments,
# so I let it be tested on Python 3.12 Linux, which is used on Ubuntu 24.04.
# On other systems, it is disabled, including mingw.

MSVC_EXPECTED = (
    # VC 2015 and 2017
    "concrt140.dll",
    "msvcp140.dll",
    "msvcp140_1.dll",
    "msvcp140_2.dll",
    "vcamp140.dll",
    "vccorlib140.dll",
    "vcomp140.dll",
    "vcruntime140.dll",
    # VS 2019
    "msvcp140_atomic_wait.dll",
    "msvcp140_codecvt_ids.dll",
    "vcruntime140_1.dll",
    # VS 2022
    "vcruntime140_threads.dll",
)

UCRT_EXPECTED = (
    "api-ms-win-*.dll",
    "ucrtbase.dll",
)


@pytest.mark.skipif(
    not (IS_WINDOWS or (IS_LINUX and sys.version_info[:2] == (3, 12))),
    reason="Windows tests",
)
def test_files() -> None:
    """Test MSVC files."""
    assert MSVC_EXPECTED == MSVC_FILES
    assert UCRT_EXPECTED == UCRT_FILES



"""Internal compatible module."""

from __future__ import annotations

import platform
import sys
from pathlib import Path
from sysconfig import get_config_var, get_platform, get_python_version

__all__ = [
    "ABI_THREAD",
    "BUILD_EXE_DIR",
    "EXE_SUFFIX",
    "EXT_SUFFIX",
    "IS_ARM_64",
    "IS_CONDA",
    "IS_LINUX",
    "IS_MACOS",
    "IS_MINGW",
    "IS_WINDOWS",
    "IS_X86_32",
    "IS_X86_64",
    "PLATFORM",
    "PYTHON_VERSION",
    "SOABI",
]

PLATFORM = get_platform()
PYTHON_VERSION = get_python_version()
ABI_THREAD = get_config_var("abi_thread") or ""

BUILD_EXE_DIR = Path(f"build/exe.{PLATFORM}-{PYTHON_VERSION}{ABI_THREAD}")
EXE_SUFFIX = get_config_var("EXE")
EXT_SUFFIX = get_config_var("EXT_SUFFIX")

IS_ARM_64 = platform.machine() in ("aarch64", "arm64", "ARM64")
IS_X86_32 = platform.machine() in ("x86", "i686")
IS_X86_64 = platform.machine() in ("x64", "x86_64", "AMD64")

IS_CONDA = Path(sys.prefix, "conda-meta").is_dir()

IS_LINUX = PLATFORM.startswith("linux")
IS_MACOS = PLATFORM.startswith("macos")
IS_MINGW = PLATFORM.startswith("mingw")
IS_WINDOWS = PLATFORM.startswith("win")

SOABI = get_config_var("SOABI")
if SOABI is None or IS_MINGW:
    # Python <= 3.12 on Windows
    # Python 3.12 MSYS2 incorrectly returns only sys.implementation.cache_tag
    platform_nodot = PLATFORM.replace(".", "").replace("-", "_")
    SOABI = f"{sys.implementation.cache_tag}-{platform_nodot}"



"""First script that is run when cx_Freeze starts up.

It determines the name of the initscript that is to be executed after a basic
initialization.
"""

from __future__ import annotations

import contextlib
import os
import string
import sys
from importlib.machinery import (
    EXTENSION_SUFFIXES,
    ExtensionFileLoader,
    ModuleSpec,
    PathFinder,
)

import BUILD_CONSTANTS

STRINGREPLACE = list(
    string.whitespace + string.punctuation.replace(".", "").replace("_", "")
)


class ExtensionFinder(PathFinder):
    """A Finder for extension modules of packages in zip files."""

    @classmethod
    def find_spec(
        cls,
        fullname,
        path=None,
        target=None,  # noqa: ARG003
    ) -> ModuleSpec | None:
        """Finder for extension modules only.

        The extension modules found within packages that are included in the
        zip file (instead of as files on disk) cannot be found within zip files
        but are stored in the lib subdirectory; if the extension module is
        found in a package, however, its name has been altered so this finder
        is needed.
        """
        if path is None:
            return None
        suffixes = EXTENSION_SUFFIXES
        for entry in sys.path:
            if not isinstance(entry, str):
                continue
            if ".zip" in entry:
                continue
            for ext in suffixes:
                location = os.path.join(entry, fullname + ext)
                if os.path.isfile(location):
                    loader = ExtensionFileLoader(fullname, location)
                    return ModuleSpec(fullname, loader, origin=location)
        return None


def get_name(executable) -> str:
    """Get the module basename to search for init and main scripts."""
    name = os.path.normcase(os.path.basename(executable))
    if sys.platform.startswith("win"):
        name, _ = os.path.splitext(name)
    name = name.partition(".")[0]
    if not name.isidentifier():
        for char in STRINGREPLACE:
            name = name.replace(char, "_")
    return name


def init() -> None:
    """Initialize the startup script."""
    # fix prefix if using console_legacy
    if not sys.prefix:
        sys.prefix = os.path.dirname(sys.executable)
        sys.base_prefix = sys.base_exec_prefix = sys.exec_prefix = sys.prefix

    # enable ExtensionFinder only if uses a zip file
    for path in sys.path:
        if path.endswith(".zip"):
            sys.meta_path.append(ExtensionFinder)
            break

    if sys.platform.startswith("win"):
        # to avoid bugs, especially in MSYS2, normalize sys.argv and sys.path
        # (preserving the reference)
        for j, path in enumerate(sys.path):
            sys.path[j] = os.path.normpath(path)
        sys.argv[0] = os.path.normpath(sys.argv[0])
        # the search path for dependencies
        search_path: list[str] = [
            entry for entry in sys.path if os.path.isdir(entry)
        ]
        # add to dll search path (or to path)
        env_path = os.environ.get("PATH", "").split(os.pathsep)
        env_path = list(map(os.path.normpath, env_path))
        for directory in search_path:
            with contextlib.suppress(OSError):
                os.add_dll_directory(directory)
            # we need to add to path for packages like 'gi' in MSYS2
            if directory not in env_path:
                env_path.insert(0, directory)
        env_path = [entry.replace(os.sep, "\\") for entry in env_path]
        os.environ["PATH"] = os.pathsep.join(env_path)
        # add extra "module.libs"
        libs_dirs = getattr(BUILD_CONSTANTS, "__LIBS__", None)
        if libs_dirs:
            for entry in libs_dirs.split(os.pathsep):
                directory = os.path.normpath(
                    os.path.join(sys.prefix, "lib", entry)
                )
                if os.path.isdir(directory):
                    os.add_dll_directory(directory)


def run() -> None:
    """Determine the name of the initscript and execute it."""
    name = get_name(sys.executable)
    try:
        # basically is __init__ plus the basename of the executable
        module_init = __import__(f"__init__{name}")
    except ModuleNotFoundError:
        # but can be renamed when only one executable exists
        executables = BUILD_CONSTANTS.__EXECUTABLES__.split(os.pathsep)
        num = len(executables)
        if num > 1:
            msg = (
                "Apparently, the original executable has been renamed to "
                f"{name!r}. When multiple executables are frozen, "
                "renaming is not allowed."
            )
            raise RuntimeError(msg) from None
        name = get_name(executables[0])
        module_init = __import__(f"__init__{name}")
    module_init.run(f"__main__{name}")



"""Compiled functions in freeze_core.util."""

from pathlib import Path

try:
    from typing import TypeAlias  # 3.10+
except ImportError:
    from typing import Any as TypeAlias

HANDLE: TypeAlias = ...

class BindError(Exception):
    """BindError Exception."""

def AddIcon(target_path: str | Path, exe_icon: str | Path) -> None:
    """Add the icon as a resource to the specified file."""

def BeginUpdateResource(
    path: str | Path, delete_existing_resources: bool = True
) -> HANDLE:
    """BeginUpdateResource wrapper."""

def UpdateResource(
    handle: HANDLE, resource_type: int, resource_id: int, resource_data: bytes
) -> None:
    """UpdateResource wrapper."""

def EndUpdateResource(handle: HANDLE, discard_changes: bool) -> None:
    """EndUpdateResource wrapper."""

def UpdateCheckSum(target_path: str | Path) -> None:
    """Update the CheckSum into the specified executable."""

def GetSystemDir() -> str:
    r"""Return the Windows system directory (C:\Windows\system for example)."""

def GetWindowsDir() -> str:
    r"""Return the Windows directory (C:\Windows for example)."""

def GetDependentFiles(path: str | Path) -> list[str]:
    """Return a list of files that this file depends on."""



"""Test winmsvcr_repack."""

from __future__ import annotations

import sys

import pytest

from freeze_core._compat import IS_LINUX, IS_MINGW, IS_WINDOWS
from freeze_core.winmsvcr import MSVC_FILES, UCRT_FILES
from freeze_core.winmsvcr_repack import copy_msvcr_files, main_test

# This test is really necessary on Windows, but it runs in other environments,
# so I let it be tested on Python 3.12 Linux, which is used on Ubuntu 24.04.
# On other systems, it is disabled, including mingw.


@pytest.mark.skipif(
    not (IS_WINDOWS or (IS_LINUX and sys.version_info[:2] == (3, 12))),
    reason="Windows tests",
)
@pytest.mark.parametrize(
    ("version", "platform", "no_cache"),
    [
        ("15", "win32", False),
        ("15", "win-amd64", False),
        ("15", "win-arm64", False),
        ("16", "win32", False),
        ("16", "win-amd64", False),
        ("16", "win-arm64", False),
        ("17", "win32", False),
        ("17", "win-amd64", True),  # just one no_cache is enough
        ("17", "win-arm64", False),
    ],
)
def test_versions(
    tmp_package, version: int, platform: str, no_cache: bool
) -> None:
    """Test the downloads of all versions of msvcr."""
    if not (IS_MINGW or IS_WINDOWS):
        tmp_package.install(["cabarchive", "striprtf"])

    copy_msvcr_files(tmp_package.path, platform, version, no_cache=no_cache)
    expected = [*MSVC_FILES]
    if version == "15":
        expected.extend(UCRT_FILES)
    names = [
        file.name.lower()
        for file in tmp_package.path.glob("*.dll")
        if any(filter(file.match, expected))
    ]
    assert names != []


@pytest.mark.skipif(
    not (IS_WINDOWS or (IS_LINUX and sys.version_info[:2] == (3, 12))),
    reason="Windows tests",
)
@pytest.mark.parametrize(
    ("version", "platform", "expected_exception", "expected_match"),
    [
        (17, "win-amd64", RuntimeError, "Version is not expected"),
        ("18", "win-amd64", RuntimeError, "Version is not expected"),
        ("17", "", RuntimeError, "Architecture not supported"),
        ("17", "win64", RuntimeError, "Architecture not supported"),
    ],
)
def test_invalid(
    tmp_package, version, platform, expected_exception, expected_match
) -> None:
    """Test invalid values to use with copy_msvcr_files function."""
    if not (IS_MINGW or IS_WINDOWS):
        tmp_package.install(["cabarchive", "striprtf"])

    with pytest.raises(expected_exception, match=expected_match):
        copy_msvcr_files(tmp_package.path, platform, version)


@pytest.mark.skipif(
    not (IS_WINDOWS or (IS_LINUX and sys.version_info[:2] == (3, 12))),
    reason="Windows tests",
)
def test_repack_main(tmp_package) -> None:
    """Test the freeze_core.winmsvcr_repack __main_ entry point with args."""
    if not (IS_MINGW or IS_WINDOWS):
        tmp_package.install(["cabarchive", "striprtf"])

    main_test(
        args=[
            f"--target-dir={tmp_package.path}",
            "--target-platform=win-amd64",
            "--version=17",
        ]
    )
    names = [
        file.name.lower()
        for file in tmp_package.path.glob("*.dll")
        if any(filter(file.match, MSVC_FILES))
    ]
    assert names != []


@pytest.mark.skipif(
    not (IS_WINDOWS or (IS_LINUX and sys.version_info[:2] == (3, 12))),
    reason="Windows tests",
)
def test_repack_main_no_option(tmp_package) -> None:
    """Test the freeze_core.winmsvcr_repack 'main' entry point without args."""
    if not (IS_MINGW or IS_WINDOWS):
        tmp_package.install(["cabarchive", "striprtf"])

    main_test(args=[])
    names = [
        file.name.lower()
        for file in tmp_package.path.glob("dist/*.dll")
        if any(filter(file.match, MSVC_FILES))
    ]
    assert names != []



"""DLL list of MSVC runtimes.

Extracted from:
    https://github.com/conda-forge/vc-feedstock/blob/main/recipe/meta.yaml

"""

from __future__ import annotations

MSVC_FILES = (
    # VC 2015 and 2017
    "concrt140.dll",
    "msvcp140.dll",
    "msvcp140_1.dll",
    "msvcp140_2.dll",
    "vcamp140.dll",
    "vccorlib140.dll",
    "vcomp140.dll",
    "vcruntime140.dll",
    # VS 2019
    "msvcp140_atomic_wait.dll",
    "msvcp140_codecvt_ids.dll",
    "vcruntime140_1.dll",
    # VS 2022
    "vcruntime140_threads.dll",
)

UCRT_FILES = (
    "api-ms-win-*.dll",
    "ucrtbase.dll",
)



"""Initialization script for cx_Freeze.

This initialization script behaves similarly to the one for console based
applications but must handle the case where Python has already been initialized
and another DLL of this kind has been loaded. As such it does not block the
path unless sys.frozen is not already set.
"""

from __future__ import annotations

import sys

if not hasattr(sys, "frozen"):
    sys.frozen = True
    sys.path = sys.path[:4]


def run() -> None:  # noqa: D103
    pass



"""Extract MSVC runtime package.

Code based on:
    https://github.com/conda-forge/vc-feedstock/blob/main/recipe/vc_repack.py
    https://github.com/conda-forge/vc-feedstock/blob/main/recipe/meta.yaml

But using cabarchive package (instead of 7z) to extract the cabs.

"""

from __future__ import annotations

import argparse
import os
import shutil
import sys
import tempfile
import time
import xml.dom.minidom
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.error import ContentTooShortError
from urllib.request import urlretrieve

from filelock import FileLock

from freeze_core._compat import PLATFORM

if TYPE_CHECKING:
    from collections.abc import Generator

__all__ = ["copy_msvcr_files", "get_msvcr_files"]

# All Microsoft CAB archives start with this signature
MS_CAB_HEADER = b"MSCF\0\0\0\0"

VC_REDIST_BASE_URL = "https://aka.ms/vs/{version}/release/{name}"

VC_VERSION_TABLE = {
    "15": "14.16.27052",
    "16": "14.29.30156",
    "17": "14.40.33816",
}

EXE_FILENAMES = {
    "win-arm64": "vc_redist.arm64.exe",
    "win-amd64": "vc_redist.x64.exe",
    "win32": "vc_redist.x86.exe",
}

VC_REDIST_CACHE = (
    Path(
        os.path.expandvars("${APPDATA}")
        if os.environ.get("APPDATA")
        else "~/.cache"
    )
    .joinpath("cxfreeze/vc_redist")
    .expanduser()
)


def split_self_extract_exe(
    exe_file: Path, target_directory: Path
) -> list[str]:
    # The self-extracting exe file contains two embedded CAB archives.
    # Split these out using a match against the known Microsoft CAB
    # header.  It's not ideal, but it works.
    contents = exe_file.read_bytes()
    splits = contents.split(MS_CAB_HEADER)
    fnames = []
    for index, s in enumerate(splits[1:]):
        fname = f"cab{index:02}.cab"
        target_directory.joinpath(fname).write_bytes(MS_CAB_HEADER + s)
        fnames.append(fname)
    return fnames


def unpack_cab(cabfile: Path, tmpdir: Path) -> None:
    import cabarchive  # noqa: PLC0415

    tmpdir.mkdir(exist_ok=True)
    cab = cabarchive.CabArchive(cabfile.read_bytes())
    for file in cab.values():
        target = tmpdir / file.filename
        target.write_bytes(file.buf)
        # Preserve timestamp
        date_time = time.mktime(
            datetime.combine(file.date, file.time).timetuple()
        )
        os.utime(target, (date_time, date_time))


def decode_manifest(directory: Path) -> dict[str, str]:
    # The first CAB file contains a manifest in the file "0" in XML format.
    dom = xml.dom.minidom.parse(os.fspath(directory / "0"))  # noqa: S318

    # The version is contained in Registration.Version
    registration = dom.documentElement.getElementsByTagName("Registration")
    version = registration[0].attributes["Version"].value
    line = f"MSVC Runtimes version: {version}"
    print("*" * len(line))
    print(line)
    print("*" * len(line))
    print(flush=True)

    # The other files have generic names such as a0, a1, etc.  The manifest
    # gives us their true names.  The FilePath contains the a0 type filename,
    # the SourcePath the true filename.
    payloads = dom.documentElement.getElementsByTagName("Payload")

    # Find the licence file.
    licences = [
        x.attributes
        for x in payloads
        if "FilePath" in x.attributes
        and x.attributes["FilePath"].value.lower() == "license.rtf"
    ]
    if len(licences) == 0:
        msg = "Found no licences in the manifest"
        raise RuntimeError(msg)
    if len(licences) > 1:
        msg = "Found more than one licence in the manifest"
        raise RuntimeError(msg)

    # Find the files that are in the second CAB file. These have
    # helpful filenames of u0, u1, etc.
    containers = [x for x in payloads if "Container" in x.attributes]

    # The DLL files we want are in the second CAB file, with a name of
    # "packages\vcRuntimeMinimum_amd64\cab1.cab",
    # "packages\VC_Runtime_arm64\cab1.cab",
    # "packages\vcRuntimeMinimum_x86\cab1.cab" or similar
    def find_cab(r, v) -> bool:
        return r in v and v.endswith(".cab")

    runtimes = [
        i.attributes
        for i in containers
        if find_cab("vcRuntimeMinimum", i.attributes["FilePath"].value)
    ]
    if len(runtimes) == 0:
        runtimes = [
            i.attributes
            for i in containers
            if find_cab("VC_Runtime", i.attributes["FilePath"].value)
        ]
    if len(runtimes) == 0:
        msg = "Found no matches in the manifest"
        raise RuntimeError(msg)
    if len(runtimes) > 1:
        msg = "Found more than one match in the manifest"
        raise RuntimeError(msg)

    return {
        "cabfile": runtimes[0]["SourcePath"].value,
        "licence": licences[0]["SourcePath"].value,
        "version": version,
    }


def fix_filename_and_copy(
    source_dir: Path, target_dir: Path
) -> Generator[Path]:
    # As of VS 17.6, the artifact contains intermediate DLL extensions,
    # which get renamed correctly upon installation; do it manually
    target_dir.mkdir(exist_ok=True)
    for file in source_dir.glob("*.dll*"):
        if file.name.startswith("api_"):
            new_fname = file.name.replace("_", "-")
        elif file.suffix.startswith(".dll_"):
            new_fname = file.stem + ".dll"
        else:
            new_fname = file.name
        print(f"Found DLL: {file.name} -> {new_fname}")
        shutil.copy2(file, target_dir / new_fname)
        yield target_dir / new_fname


def unpack_exe(exe_filename: Path, unpack_dir: Path) -> Generator[Path]:
    from striprtf.striprtf import rtf_to_text  # noqa: PLC0415

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        cabs = split_self_extract_exe(exe_filename, tmp_path)
        # The first cab is the installer data
        # The second is the payload
        cabfile1 = tmp_path / cabs[0]
        cabdir1 = tmp_path / cabfile1.stem
        unpack_cab(cabfile1, cabdir1)
        payload = decode_manifest(cabdir1)
        # Get LICENSE and converts to txt
        unpack_dir.mkdir(parents=True, exist_ok=True)
        license_rtf = unpack_dir / "LICENSE.RTF"
        shutil.copy2(cabdir1 / payload["licence"], license_rtf)
        yield license_rtf
        text = rtf_to_text(
            license_rtf.read_text(encoding="cp1252", errors="strict")
        )
        license_txt = unpack_dir / "LICENSE.txt"
        license_txt.write_text(text, encoding="cp1252", errors="strict")
        yield license_txt
        # Get the payload
        cabfile2 = tmp_path / cabs[1]
        cabdir2 = tmp_path / cabfile2.stem
        unpack_cab(cabfile2, cabdir2)
        vc_redist = tmp_path / exe_filename.stem
        vc_redist.mkdir(exist_ok=True)
        unpack_cab(Path(cabdir2, payload["cabfile"]), vc_redist)
        yield from fix_filename_and_copy(vc_redist, unpack_dir)


def get_msvcr_files(
    version: str | None = None,
    target_platform: str | None = None,
    no_cache: bool = False,
) -> Generator[Path]:
    """Get MSVC runtime files."""
    if target_platform is None:
        target_platform = (
            PLATFORM if PLATFORM.startswith("win") else "win-amd64"
        )
    if target_platform in EXE_FILENAMES:
        name = EXE_FILENAMES[target_platform]
    else:
        msg = f"Architecture not supported: {target_platform!r}"
        raise RuntimeError(msg)
    if version is None:
        version = max(VC_VERSION_TABLE.keys())
    if version not in VC_VERSION_TABLE:
        msg = f"Version is not expected: {version!r}"
        raise RuntimeError(msg)

    # use a cache dir
    cache_dir = VC_REDIST_CACHE / version
    cache_dir.mkdir(parents=True, exist_ok=True)
    filename = cache_dir / name
    unpack_dir = cache_dir / name.replace(".", "_")
    unpack_dir.mkdir(parents=True, exist_ok=True)
    with FileLock(filename.with_suffix(".lock")):
        if no_cache or (filename.exists() and filename.stat().st_size == 0):
            filename.unlink(missing_ok=True)
        url = VC_REDIST_BASE_URL.format(version=version, name=name)
        try:
            while not filename.exists():
                urlretrieve(url, filename)  # noqa: S310
        except ContentTooShortError as exc:
            print("warning:", exc.reason, "of", name, file=sys.stderr)
            print("retry!", file=sys.stderr)
            filename.unlink(missing_ok=True)
        if filename.exists():
            files = list(unpack_dir.glob("*.dll"))
            if no_cache or not files:
                yield from unpack_exe(filename, unpack_dir)
            else:
                yield from unpack_dir.iterdir()
        else:
            msg = f"File not found: {filename}"
            raise RuntimeError(msg)


def copy_msvcr_files(
    target_dir: Path | str,
    target_platform: str | None = None,
    version: str | None = None,
    dry_run: bool = False,
    no_cache: bool = False,
) -> None:
    """Copy MSVC runtime files."""
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    for file in get_msvcr_files(version, target_platform, no_cache):
        if not dry_run:
            shutil.copy2(file, target_dir / file.name)


def main_test(args=None) -> None:
    """Command line test."""
    parser = argparse.ArgumentParser(
        description="Extract MSVC runtime package"
    )
    parser.add_argument(
        "--dry-run",
        "-n",
        help="Do not copy files, list files only",
        action="store_true",
    )
    parser.add_argument(
        "--target-dir", help="dist", metavar="DIR", default="dist"
    )
    parser.add_argument(
        "--target-platform",
        help="Target architecture (eg. win-arm64, win-amd64, win32)",
        default=None,
    )
    parser.add_argument(
        "--version",
        help="Runtime version number",
        default=None,
    )
    parser.add_argument(
        "--no-cache",
        help="Don't use the cached runtime",
        action="store_true",
    )
    args = parser.parse_args(args)

    copy_msvcr_files(
        args.target_dir,
        args.target_platform,
        args.version,
        args.dry_run,
        args.no_cache,
    )


if __name__ == "__main__":
    main_test()



"""The source of the base executables.

The base executables are used to launch your Python applications. Different
bases serve for different types of application on Windows
(GUI, console application or service).
The base executable calls the initscript, which in turn calls the user's code.
"""
